/*
Copyright (C) 1996-2001  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/xpm.h>

#include "MwCanvasP.h"

#define DEFAULT_HEIGHT 50
#define DEFAULT_WIDTH 50

#define offset(field) XtOffsetOf(MwCanvasRec, canvas.field)
static XtResource resources[] = {
	{
#if 0
		XtNcanvasExpose,	/* name */
		XtCCanvasExpose,	/* class */
		XtRPointer,		/* type */
		sizeof(XtPointer),	/* size */
		offset(expose),		/* offset */
		XtRImmediate,		/* default_type */
		(XtPointer) NULL	/* default_addr */
#else
		XtNcallback,
		XtCCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(callbacks),
		XtRCallback,
		(XtPointer)NULL
#endif
#if 0
	}, {
		XtNcanvasData,
		XtCCanvasData,
		XtRPointer,
		sizeof(XtPointer),
		offset(data),
		XtRImmediate,
		(XtPointer) NULL
#endif
	}
};
#undef offset

/* methods */
static void Redisplay(Widget, XEvent *, Region);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);
static void Initialize(Widget, Widget, ArgList, Cardinal *);

/* actions */
static void MwCanvasAction(Widget, XEvent *, String *, Cardinal *);

static XtActionsRec actions[] =
{
	{"canvas", MwCanvasAction},
};

/* translations */
static char translations[] =
"<Key>:		canvas()\n";

#define SuperClass ((CoreWidgetClass)&coreClassRec)

MwCanvasClassRec mwCanvasClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) SuperClass,
    /* class_name		*/	"MwCanvas",
    /* widget_size		*/	sizeof(MwCanvasRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* canvas fields */
    /* empty			*/	0
  }
};

WidgetClass mwCanvasWidgetClass = (WidgetClass)&mwCanvasClassRec;


/* supporting code copied directly from canvas.c */

static void MwCanvasAction(Widget w, XEvent *event, String *params, Cardinal *n)
{
	;
}

static void Initialize(Widget req, Widget new, ArgList args, Cardinal *nargs)
{
	MwCanvasWidget cw = (MwCanvasWidget)new;

printf("Initialize(%s)\n", XtName(new));
	if (cw->core.height == 0) cw->core.height = DEFAULT_HEIGHT;
	if (cw->core.width == 0) cw->core.width = DEFAULT_WIDTH;

/*
widgetClassRec.core_class.resize((Widget)cw);
	(*XtClass(new)->core_class.resize)((Widget)cw);
*/
}

static void Redisplay(Widget w, XEvent *event, Region r)
{
	MwCanvasWidget aw = (MwCanvasWidget) w;

#if 0
	if (aw->canvas.expose)
		(*aw->canvas.expose)(w, aw->canvas.data);
#else
	XtCallCallbackList(w, aw->canvas.callbacks, NULL);
#endif
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *nargs)
{
	Boolean do_redisplay = True;

	return do_redisplay;
}

