/* 

Copyright (c) 1996-2001  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>

#ifdef HAVE_FNMATCH
#include <fnmatch.h>
#else
#ifdef HAVE_GMATCH
#include <libgen.h>
#define fnmatch(a,b,c) (!gmatch(b,a))
#else
#define fnmatch(a,b,c) 0                /* match anything */
#endif
#endif

#include <sys/stat.h>
#include <sys/param.h>
#include <unistd.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "Mowitz.h"
#include "MwFileselP.h"
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>

#define offset(field) XtOffsetOf(MwFileselRec, filesel.field)

static XtResource resources[] = {
	{
		XtNformat,
		XtCFormat,
		XtRString,
		sizeof(String),
		offset(format),
		XtRString,
		(XtPointer)NULL
	},
	{
		XtNcancelCallback,
		XtCCancelCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(cancel_callbacks),
		XtRCallback,
		(XtPointer)NULL
	},
	{
		XtNselectCallback,
		XtCSelectCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(select_callbacks),
		XtRCallback,
		(XtPointer)NULL
	},
	{
		XtNdirLabel,
		XtCLabel,
		XtRString,
		sizeof(String),
		offset(dir_label),
		XtRString,
		"Directory:"
	},
	{
		XtNxLayout,
		XtCXLayout,
		XtRString,
		sizeof(String),
		XtOffsetOf(MwFileselRec, rudegrid.x_layout),
		XtRString,
		"5 100 5 50% 5 50% 5 100 5"
	},
	{
		XtNyLayout,
		XtCYLayout,
		XtRString,
		sizeof(String),
		XtOffsetOf(MwFileselRec, rudegrid.y_layout),
		XtRString,
		"35 5 25 5 25 5 100% 5 22 5 32"
	}
};

#undef offset

static void ClassInitialize();
static void Initialize();
static void Destroy();
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

#define SuperClass ((WidgetClass)&mwRudegridClassRec)

MwFileselClassRec mwFileselClassRec = {
	{	/* core_class fields */
		SuperClass,			/* superclass */
		"MwFilesel",			/* class_name */
		sizeof(MwFileselRec),		/* widget_size */
		ClassInitialize,		/* class_initialize */
		NULL,				/* class_part_init */
		FALSE,				/* class_inited */
		Initialize,			/* initialize */
		NULL,				/* initialize_hook */
		XtInheritRealize,		/* realize */
		NULL,				/* actions */
		0,				/* num_actions */
		resources,			/* resources */
		XtNumber(resources),		/* num_resources */
		NULLQUARK,			/* xrm_class */
		TRUE,				/* compress_motion */
		TRUE,				/* compress_exposure */
		TRUE,				/* compress_enterleave */
		FALSE,				/* visible_interest */
		Destroy,			/* destroy */
		XtInheritResize,		/* resize */
		XtInheritExpose,		/* expose */
		SetValues,			/* set_values */
		NULL,				/* set_values_hook */
		XtInheritSetValuesAlmost,	/* set_values_almost */
		NULL,				/* get_values_hook */
		NULL,				/* accept_focus */
		XtVersion,			/* version */
		NULL,				/* callback_private */
		NULL,				/* tm_table */
		XtInheritQueryGeometry,		/* query_geometry */
		XtInheritDisplayAccelerator,	/* display_accelerator */
		NULL,				/* extension */
	},
	{	/* composite_class fields */
		XtInheritGeometryManager,	/* geometry_manager */
		XtInheritChangeManaged,		/* change_managed */
		XtInheritInsertChild,		/* insert_child */
		XtInheritDeleteChild,		/* delete_child */
		NULL,				/* extension */
	},
	{	/* constraint_class fields */
		NULL,				/* subresources */
		0,				/* subresources_count */
		sizeof(MwFileselConstraintsRec),/* constraint_size */
		NULL,				/* initialize */
		NULL,				/* destroy */
		NULL,				/* set_values */
		NULL,				/* extension */
	},
	{	/* rudegrid_class fields */
		NULL,				/* extension */
	},
	{	/* filesel_class fields */
		NULL,				/* extension */
	}
};

WidgetClass mwFileselWidgetClass = (WidgetClass)&mwFileselClassRec;

/* Utility functions stolen from filesel.c in Siag Office 3.4.10 */

static MwFileselWidget find_filesel(Widget w)
{
printf("find_filesel(%s)", XtName(w));
	while (w != None && XtClass(w) != mwFileselWidgetClass)
		w = XtParent(w);
printf(" => %s\n", w?XtName(w):"None");
	return (MwFileselWidget)w;
}

static int compar(const void *p, const void *q)
{
	return strcmp(*(const char **)p, *(const char **)q);
}

static int
getdirent(char ***ldirp, int *ndirp, char ***lfilep, int *nfilep,
        char *path, char *pattern)
{
        DIR *dird;
        struct dirent *dire;
        char fn[1024];
        struct stat buf;
        char **ldir = NULL, **lfile = NULL;
        int ndir = 0, mdir = 0, nfile = 0, mfile = 0;

        if ((dird = opendir(path)) == NULL) {
                fprintf(stderr, "Can't open %s\n", path);
                return 1;
        }
        while ((dire = readdir(dird))) {
                sprintf(fn, "%s/%s", path, dire->d_name);
                if ((stat(fn, &buf))) {
                        fprintf(stderr, "Can't stat %s\n", fn);
                        continue;
                }
                if (S_IFDIR & buf.st_mode) {
                        if (ndir >= mdir)
                                ldir = MwRealloc(ldir,
                                        (mdir += 256)*sizeof(char *));
                        ldir[ndir++] = MwStrdup(dire->d_name);
                }
                else if (S_IFREG & buf.st_mode && 
                                !fnmatch(pattern, dire->d_name, 0)) {
                        if (nfile >= mfile)
                                lfile = MwRealloc(lfile,
                                        (mfile += 256)*sizeof(char *));
                        lfile[nfile++] = MwStrdup(dire->d_name);
                }
        }
        closedir(dird);

        qsort(ldir, ndir, sizeof(char *), compar);
        qsort(lfile, nfile, sizeof(char *), compar);

        *ndirp = ndir;
        *ldirp = ldir;
        *nfilep = nfile;
        *lfilep = lfile;

        return 0;
}

static void
freedirent(char **ldir, int ndir, char **lfile, int nfile)
{
        int i;

        for (i = 0; i < ndir; i++)
                MwFree(ldir[i]);
        MwFree(ldir);
        for (i = 0; i < nfile; i++)
                MwFree(lfile[i]);
        MwFree(lfile);
}

/* and that's all there is to it */


/* Foolproof default format */
static char *fileformats[] = {
        "All (*)",
        NULL};


static void make_files(MwFileselWidget fw, String files[], Cardinal nfiles)
{
        static String foo[1] = {"<none>"};

        if (nfiles == 0)
                XawListChange(fw->filesel.filelist,
                        foo, 1, 0, True);
        else
                XawListChange(fw->filesel.filelist,
                        files, nfiles, 0, True);
}

static void make_dirs(MwFileselWidget fw, String dirs[], Cardinal ndirs)
{
        XawListChange(fw->filesel.dirlist,
                dirs, ndirs, 0, True);
}

static void make_dirmenu(MwFileselWidget, char *);

static void fsel_scan(MwFileselWidget fw)
{
        String string;
        char dir[1024];
        char pattern[1024];
        char *pst;

printf("fsel_scan(%s)\n", XtName((Widget)fw));
        if (fw->filesel.files != NULL)
                freedirent(fw->filesel.dirs, fw->filesel.ndirs,
			   fw->filesel.files, fw->filesel.nfiles);
        
        strcpy(dir, MwLabelGet(fw->filesel.dirbutton));
        string = MwLabelGet(fw->filesel.formatbutton);
        pst = strchr(string, '(');
        if (pst == NULL) {
                /* apparently no pattern here; use foolproof default */
                strcpy(pattern, "*");
        } else {
                strcpy(pattern, pst+1);
                if ((pst = strchr(pattern, ')')))
                        *pst = '\0';
        }
printf("getdirent with dir = '%s', pattern = '%s'\n", dir, pattern);
        if (getdirent(&fw->filesel.dirs, &fw->filesel.ndirs,
		      &fw->filesel.files, &fw->filesel.nfiles, dir, pattern)) {
                fprintf(stderr, "Grmbl. getdirent() failed\n");
        }
        make_files(fw, fw->filesel.files, fw->filesel.nfiles);
        make_dirs(fw, fw->filesel.dirs, fw->filesel.ndirs);
        make_dirmenu(fw, dir);
}

/* ---
A format pattern looks like this:
        Plaintext description of foo format (*.foo)
We want to return .foo
*/
static char *get_extension(char *f, char *x)
{
        int n;
        char *p = strrchr(f, '(');
        if (p == NULL) return NULL;
        p = strchr(p, '.');
        if (p == NULL) return NULL;
        strcpy(x, p);
        n = strcspn(x, ")[]*");
        x[n] = '\0';
        return x;
}

static void format_select(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwFileselWidget fw = find_filesel(w);
        static char filename[1024];
        char *p, *x, ext[100];
        strncpy(filename, (char *)client_data, sizeof filename);

        if (fw->filesel.ch_ext) {
                p = MwTextFieldGetString(fw->filesel.filetext);
                strncpy(filename, p, sizeof filename);
                p = strrchr(filename, '.');
                x = get_extension((char *)client_data, ext);
                if (p && x) {
                        strcpy(p, x);
                        XtVaSetValues(fw->filesel.filetext,
                                XtNstring, filename,
                                (char *)0);
                }
        }
        MwLabelSet(fw->filesel.formatbutton, (String)client_data);
	MwFree(fw->filesel.format);
	fw->filesel.format = MwStrdup((char *)client_data);
        fsel_scan(fw);
}

static void file_select(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwFileselWidget fw = find_filesel(w);
        XawListReturnStruct *list_struct = (XawListReturnStruct *)call_data;
        String string = list_struct->string;

        XtVaSetValues(fw->filesel.filetext,
                XtNstring, string, (char *)0);
        XawListUnhighlight(fw->filesel.filelist);
        XawListUnhighlight(fw->filesel.dirlist);
}

static void dir_select(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwFileselWidget fw = find_filesel(w);
        char path[1024];
        char newpath[1024];

        XawListReturnStruct *list_struct = (XawListReturnStruct *)call_data;
        String string = list_struct->string;

        sprintf(path, "%s/%s", MwLabelGet(fw->filesel.dirbutton), string);

        if (!realpath(path, newpath))
                fprintf(stderr, "Couldn't realpath %s\n", path);

        MwLabelSet(fw->filesel.dirbutton, newpath);

        XawListUnhighlight(fw->filesel.filelist);
        XawListUnhighlight(fw->filesel.dirlist);
        fsel_scan(fw);
}

static void make_menu(char **menu_entry_names, Widget menu)
{
        int i;
        Widget entry;

        for (i = 0; menu_entry_names[i]; i++) {
                String item = menu_entry_names[i];

                if (item[0] == '-')     /* line pane */
                        entry = XtVaCreateManagedWidget(item,
                                mwLineMEObjectClass, menu,
                                (char *)0);
                else {
                        entry = XtVaCreateManagedWidget(item,
                                mwLabelMEObjectClass, menu,
                                XtNlabel, item,
                                (char *)0);
                        XtAddCallback(entry,
                                XtNcallback, format_select,
                                menu_entry_names[i]);
                }
        }
}

static void change_dir(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwFileselWidget fw = find_filesel(w);
        char path[1024];
        char newpath[1024];

        strcpy(path, (char *)client_data);

        if (!realpath(path, newpath))
                fprintf(stderr, "Couldn't realpath %s\n", path);

        MwLabelSet(fw->filesel.dirbutton, newpath);

        fsel_scan(fw);
}

static void make_dirmenu(MwFileselWidget fw, char *dir)
{
        int i;
        char *p, *b, *item;

        b = MwStrdup(dir);
        for (i = 0; i < fw->filesel.nentry; i++) {
                if (fw->filesel.dir_entry[i].w != None) {
                        XtDestroyWidget(fw->filesel.dir_entry[i].w);
                        MwFree(fw->filesel.dir_entry[i].d);
                }
        }
        fw->filesel.nentry = 0;

        XtVaSetValues(fw->filesel.dirbutton,
                XtNwidth, 200, (char *)0);
        if (fw->filesel.dirmenu != None)
                XtDestroyWidget(fw->filesel.dirmenu);
        fw->filesel.dirmenu = XtVaCreatePopupShell("dirmenu",
                mwMenuWidgetClass, XtParent(fw->filesel.dirbutton),
                (char *)0);

        while ((p = strrchr(b, '/'))) {
                *p = '\0';
                if (p == b)     /* root dir */
                        item = MwStrdup("/");
                else
                        item = MwStrdup(b);
                fw->filesel.dir_entry[fw->filesel.nentry].d = item;

                fw->filesel.dir_entry[fw->filesel.nentry].w = XtVaCreateManagedWidget(item,
                        mwLabelMEObjectClass, fw->filesel.dirmenu,
                        XtNlabel, item,
                        (char *)0);
                XtAddCallback(fw->filesel.dir_entry[fw->filesel.nentry].w,
                        XtNcallback, change_dir,
                        fw->filesel.dir_entry[fw->filesel.nentry].d);
                fw->filesel.nentry++;
        }
}

static void do_select_callbacks(Widget w)
{
	MwFileselWidget fw = find_filesel(w);
	char path[1024];
	String string;
printf("do_select_callbacks(%s)\n", XtName(w));
	XtVaGetValues(fw->filesel.filetext,
		XtNstring, &string, (char *)0);
	sprintf(path, "%s/%s", MwLabelGet(fw->filesel.dirbutton), string);
	XtCallCallbackList((Widget)fw, fw->filesel.select_callbacks,
		(XtPointer)path);
}

static void do_cancel_callbacks(Widget w)
{
	MwFileselWidget fw = find_filesel(w);
printf("do_cancel_callbacks(%s)\n", XtName(w));
	XtCallCallbackList((Widget)fw, fw->filesel.cancel_callbacks,
		(XtPointer)NULL);
}

static void fsel_done_action(Widget w, XEvent *event, String *params, Cardinal *
n)
{
	MwFileselWidget fw = find_filesel(w);
	String string;
	char path[1024], newpath[1024];
	struct stat buf;
printf("fsel_done_action(%s)\n", XtName(w));
        XtVaGetValues(fw->filesel.filetext,
                XtNstring, &string, (char *)0);
        if (string[0] == '/') strcpy(path, string);
        else sprintf(path, "%s/%s", MwLabelGet(fw->filesel.dirbutton), string);
        realpath(path, newpath);
        if (!stat(newpath, &buf) && (S_IFDIR & buf.st_mode)) {
                MwLabelSet(fw->filesel.dirbutton, newpath);
                XtVaSetValues(fw->filesel.filetext,
                        XtNstring, "", (char *)0);
                fsel_scan(fw);
        } else {
        	do_select_callbacks(w);
        }
}

static void fsel_done(Widget w, XtPointer client_data, XtPointer call_data)
{
printf("fsel_done(%s)\n", XtName(w));
	do_select_callbacks(w);
}

static void fsel_cancel_action(Widget w, XEvent *event,
                String *params, Cardinal *n)
{
printf("fsel_cancel_action(%s)\n", XtName(w));
	do_cancel_callbacks(w);
}

static void fsel_abort(Widget w, XtPointer client_data, XtPointer call_data)
{
printf("fsel_abort(%s)\n", XtName(w));
	do_cancel_callbacks(w);
}


static XtActionsRec actions[] =
{
        {"fsel-done", fsel_done_action},
        {"fsel-cancel", fsel_cancel_action}
};

/* ---
*/
static Widget add_button(char *name, char *label,
        void (*cb)(Widget, XtPointer, XtPointer), XtPointer data, Widget pw)
{
        Widget below = XtVaCreateManagedWidget(name,
                commandWidgetClass, pw,
                XtNwidth, 80,
                (char *)0);
        MwLabelSet(below, label);
        XtVaSetValues(below, XtNwidth, 80, (char *)0);
        if (cb) XtAddCallback(below, XtNcallback, cb, data);
        return below;
}

static void fsel_find(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwFileselWidget fw = find_filesel(w);
        char fn[1024];
        char cmd[1024]; 
        char **files = NULL;
        int nfiles = 0, i;
        FILE *fp;

	if (w/*fsel_pshell*/ == None) return;

        fn[0] = '\0';
        if (MwDialogInput(w/*fsel_pshell*/, "Find pattern:", fn)) {
                if (fn[0] == '!') {
                        sprintf(cmd, "find %s %s -print",
                                MwLabelGet(fw->filesel.dirbutton), fn+1);
                } else {
                        sprintf(cmd, "find %s -name '%s' -print",
                                MwLabelGet(fw->filesel.dirbutton), fn);
                }
                if (!(fp = popen(cmd, "r"))) {
                        MwErrorBox(w/*fsel_pshell*/, "Can't exec find");
                        return;
                }
                while ((fgets(fn, sizeof fn, fp))) {
                        MwChomp(fn);
                        files = MwRealloc(files, (nfiles+1)*(sizeof(char *)));
                        files[nfiles++] = MwStrdup(fn);
                }
                pclose(fp);
                if (nfiles) {
                        int n = MwListBox(w/*fsel_pshell*/,
                                "Pick one:", files, nfiles);
                        if (n != -1) {
                                char *p = strrchr(files[n], '/');
                                if (p) {
                                        *p = '\0';
                                        XtVaSetValues(fw->filesel.filetext,
                                                XtNstring, p+1, (char *)0);
                                        change_dir(w,
                                                (XtPointer)files[n], NULL);
                                } else {
                                        XtVaSetValues(fw->filesel.filetext,
                                                XtNstring, files[n], (char *)0);
                                }
                        }
                } else {
                        MwErrorBox(w/*fsel_pshell*/, "No files found");
                }
                for (i = 0; i < nfiles; i++)
                        MwFree(files[i]);
                MwFree(files);
        }
}

static void fsel_help(Widget w, XtPointer client_data, XtPointer call_data)
{
	MwHelp("Filesel.html");
}

static Atom wm_delete_window;


static void ClassInitialize(void)
{
	;
}

static void Initialize(Widget req, Widget new, ArgList args, Cardinal *nargs)
{
	MwFileselWidget fw = (MwFileselWidget)new;

	XtAppContext ac = XtWidgetToApplicationContext(new);
	XtAppAddActions(ac, actions, XtNumber(actions));

	if (fw->core.width < 400) fw->core.width = 400;
	if (fw->core.height < 400) fw->core.height = 400;

	fw->filesel.topframe = XtVaCreateManagedWidget("topframe",
		mwFrameWidgetClass, new,
		XtNshadowType, Raised,
		XtNgridWidth, 9,
		XtNshadowWidth, 1,
		(char *)0);
	fw->filesel.topbox = XtVaCreateManagedWidget("topbox",
		boxWidgetClass, fw->filesel.topframe,
		XtNborderWidth, 0,
		(char *)0);
	fw->filesel.bottomframe = XtVaCreateManagedWidget("bottomframe",
		mwFrameWidgetClass, new,
		XtNshadowType, Raised,
		XtNgridy, 10,
		XtNgridWidth, 9,
		XtNshadowWidth, 1,
		(char *)0);
	fw->filesel.bottombox = XtVaCreateManagedWidget("bottombox",
		boxWidgetClass, fw->filesel.bottomframe,
		XtNborderWidth, 0,
		(char *)0);
	fw->filesel.filelabel = XtVaCreateManagedWidget("filelabel",
		labelWidgetClass, new,
		XtNjustify, XtJustifyLeft,
		XtNgridx, 1,
		XtNgridy, 8,
		(char *)0);
	MwLabelSet(fw->filesel.filelabel, "File Name:");
	fw->filesel.textframe = XtVaCreateManagedWidget("textframe",
		mwFrameWidgetClass, new,
		XtNgridx, 3,
		XtNgridy, 8,
		XtNgridWidth, 5,
		XtNallowResize, False,
		XtNshadowType, Lowered,
		XtNshadowWidth, 1,
		(char *)0);
	fw->filesel.filetext = XtVaCreateManagedWidget("filetext",
		mwTextfieldWidgetClass, fw->filesel.textframe,
		XtNborderWidth, 0,
		(char *)0);
	XtVaSetValues(fw->filesel.filetext, XtNstring, "", (char *)0);
	XtOverrideTranslations(fw->filesel.filetext,
		XtParseTranslationTable(
			"<Key>Return:	fsel-done()\n"
			"<Key>Escape:	fsel-cancel()"));

	fw->filesel.fileframe = XtVaCreateManagedWidget("fileframe",
		mwFrameWidgetClass, new,
		XtNgridx, 5,
		XtNgridy, 6,
		XtNgridWidth, 3,
		XtNshadowType, Groove,
		XtNmarginWidth, 2,
		XtNmarginHeight, 2,
		XtNallowResize, False,
		(char *)0);
	fw->filesel.filelabel2 = XtVaCreateManagedWidget("filelabel2",
		labelWidgetClass, fw->filesel.fileframe,
		(char *)0);
	MwLabelSet(fw->filesel.filelabel2, "Files");
	XtVaSetValues(fw->filesel.fileframe,
		XtNtitle, fw->filesel.filelabel2,
		(char *)0);
	fw->filesel.fileviewport = XtVaCreateManagedWidget("fileviewport",
		viewportWidgetClass, fw->filesel.fileframe,
		XtNforceBars, True,
		XtNborderWidth, 0,
		XtNallowHoriz, True,
		XtNallowVert, True,
		XtNuseBottom, True,
		XtNuseRight, True,
		(char *)0);
	fw->filesel.filelist = XtVaCreateManagedWidget("filelist",
		listWidgetClass, fw->filesel.fileviewport,
		XtNdefaultColumns, 1,
		XtNforceColumns, 1,
		(char *)0);
	XtAddCallback(fw->filesel.filelist, XtNcallback, file_select, NULL);

	fw->filesel.formatlabel = XtVaCreateManagedWidget("formatlabel",
		labelWidgetClass, new,
		XtNjustify, XtJustifyLeft,
		XtNgridx, 1,
		XtNgridy, 2,
		(char *)0);
	MwLabelSet(fw->filesel.formatlabel, "Format:");

	fw->filesel.formatbutton = XtVaCreateManagedWidget("formatbutton",
		mwMenuButtonWidgetClass, new,
		XtNmenu_name, "formatmenu",
		XtNgridx, 3,
		XtNgridy, 2,
		XtNgridWidth, 5,
		XtNjustify, XtJustifyLeft,
		XtNshadowWidth, 1,
		(char *)0);
	fw->filesel.dirlabel = XtVaCreateManagedWidget("dirlabel",
		labelWidgetClass, new,
		XtNjustify, XtJustifyLeft,
		XtNgridx, 1,
		XtNgridy, 4,
		(char *)0);
	MwLabelSet(fw->filesel.dirlabel, "Directory:");

	fw->filesel.dirbutton = XtVaCreateManagedWidget("dirbutton",
		mwMenuButtonWidgetClass, new,
		XtNmenu_name, "dirmenu",
		XtNgridx, 3,
		XtNgridy, 4,
		XtNgridWidth, 5,
		XtNjustify, XtJustifyLeft,
		XtNshadowWidth, 1,
		(char *)0);
	fw->filesel.dirframe = XtVaCreateManagedWidget("dirframe",
		mwFrameWidgetClass, new,
		XtNgridx, 1,
		XtNgridy, 6,
		XtNgridWidth, 3,
		XtNshadowType, Groove,
		XtNmarginWidth, 2,
		XtNmarginHeight, 2,
		XtNallowResize, False,
		(char *)0);
	fw->filesel.dirlabel2 = XtVaCreateManagedWidget("dirlabel2",
		labelWidgetClass, fw->filesel.dirframe,
		(char *)0);
	MwLabelSet(fw->filesel.dirlabel2, "Directories");
	XtVaSetValues(fw->filesel.dirframe,
		XtNtitle, fw->filesel.dirlabel2,
		(char *)0);
	fw->filesel.dirviewport = XtVaCreateManagedWidget("dirviewport",
		viewportWidgetClass, fw->filesel.dirframe,
		XtNforceBars, True,
		XtNborderWidth, 0,
		XtNallowHoriz, True,
		XtNallowVert, True,
		XtNuseBottom, True,
		XtNuseRight, True,
		(char *)0);
	fw->filesel.dirlist = XtVaCreateManagedWidget("dirlist",
		listWidgetClass, fw->filesel.dirviewport,
		XtNdefaultColumns, 1,
		XtNforceColumns, 1,
		(char *)0);
	XtAddCallback(fw->filesel.dirlist, XtNcallback, dir_select, NULL);

	fw->filesel.okbutton = add_button("fsel_okbutton", "OK",
		fsel_done, NULL, fw->filesel.bottombox);
	fw->filesel.cancelbutton = add_button("fsel_cancelbutton", "Cancel",
		fsel_abort, NULL, fw->filesel.bottombox);
	fw->filesel.findbutton = add_button("fsel_findbutton", "Find",
		fsel_find, NULL, fw->filesel.topbox);
	fw->filesel.helpbutton = add_button("fsel_helpbutton", "Help",
		fsel_help, NULL, fw->filesel.topbox);
}

static void Destroy(Widget w)
{
	;
}

static Boolean SetValues(Widget cur, Widget req, Widget new,
	ArgList args, Cardinal *nargs)
{
	MwFileselWidget newfw = (MwFileselWidget)new;
	MwFileselWidget curfw = (MwFileselWidget)cur;
	if (curfw->filesel.format != newfw->filesel.format) {
		MwFree(curfw->filesel.format);
		newfw->filesel.format = MwStrdup(newfw->filesel.format);
		MwLabelSet(newfw->filesel.formatbutton,
			newfw->filesel.format);
	}
	return False;
}
	
/*
A convenience function which creates a file selector widget
in a new shell.
*/

static Widget fsel_pshell;
static char *fsel_path, *fsel_fmt;
static int fsel_status;

static void input_done(Widget w, XtPointer client_data, XtPointer call_data)
{
	String fmt;
printf("input_done(%s)\n", XtName(w));
	strcpy(fsel_path, (char *)call_data);
	XtVaGetValues(w, XtNformat, &fmt, (char *)0);
	strcpy(fsel_fmt, fmt);
printf("path = '%s'\n", fsel_path);
        XtPopdown(fsel_pshell);
        fsel_status = MW_DONE;
}

static void input_abort(Widget w, XtPointer client_data, XtPointer call_data)
{
printf("input_abort(%s)\n", XtName(w));
        XtPopdown(fsel_pshell);
        fsel_status = MW_ABORT;
}

static void filesel_close_action(Widget w, XEvent *event,
		String *params, Cardinal *n)
{
printf("filesel_close_action(%s)\n", XtName(w));
	input_abort(w, NULL, NULL);
}

static XtActionsRec input_actions[] =
{
	{"filesel-close", filesel_close_action}
};

static MwFileselWidget fsel_init(Widget topLevel)
{
	Widget fsel_form;
        XtAppContext app_context = XtWidgetToApplicationContext(topLevel);

	XtAppAddActions(app_context, input_actions, XtNumber(input_actions));
        fsel_pshell = XtVaCreatePopupShell("fsel_pshell",
                transientShellWidgetClass, topLevel,
                XtNtitle, _("Select File"),
                (char *)0);

        fsel_form = XtCreateManagedWidget("fsel_form",
                mwFileselWidgetClass, fsel_pshell, NULL, 0);
	XtAddCallback(fsel_form, XtNcancelCallback,
		input_abort, NULL);
	XtAddCallback(fsel_form, XtNselectCallback,
		input_done, NULL);

	/* make the widget call input_abort if the shell is closed */
        wm_delete_window = XInternAtom(XtDisplay(fsel_pshell),
                                "WM_DELETE_WINDOW", False);
        XtOverrideTranslations(fsel_pshell,
                XtParseTranslationTable(
                        "<Message>WM_PROTOCOLS: filesel-close()"));
	return (MwFileselWidget)fsel_form;
}

static void add_extra(MwFileselWidget fw, char *extra)
{
        char *p, *q;

        fw->filesel.nextra = 0;
        if (!extra) return;
        for (p = strtok(extra, ":"); p; p = strtok(NULL, ":")) {
                q = strchr(p, '=');
                if (!q) continue;
                *q++ = '\0';
                fw->filesel.extra[fw->filesel.nextra] = add_button("extra", p,
                        change_dir, q, fw->filesel.topbox);
                XtVaSetValues(fw->filesel.extra[fw->filesel.nextra++],
                        XtNwidth, 80, (char *)0);
        }
}

/* ---
Select a file and directory. The path and name must contain valid strings.
The format is returned in fmt. Extra is a list of extra directory buttons
to appear on the top row. If ext != 0, change extension of file name
when format pattern is changed.
*/
int MwFileselInput(Widget pw, char *path, char *name, char **patterns,
        char *fmt, char *extra, int ext)
{
	char newpath[1024];
        XtAppContext app_context = XtWidgetToApplicationContext(pw);
        int i;
	static MwFileselWidget fw;

        if (fsel_pshell == None) fw = fsel_init(pw);
	fsel_path = path;
	fsel_fmt = fmt;

        if (patterns == NULL)
                patterns = fileformats;

        /* First, create all the formats on the menu */
        XtVaSetValues(fw->filesel.formatbutton,
                XtNlabel, patterns[0],
                XtNwidth, 200, (char *)0);
	fw->filesel.format = MwStrdup(patterns[0]);
        fw->filesel.formatmenu = XtVaCreatePopupShell("formatmenu",
                mwMenuWidgetClass, XtParent(fw->filesel.formatbutton),
                (char *)0);

        make_menu(patterns, fw->filesel.formatmenu);
        fw->filesel.ch_ext = ext;

        if (!realpath(path, newpath))
                fprintf(stderr, "Couldn't realpath %s\n", path);

        MwLabelSet(fw->filesel.dirbutton, newpath);

        XtVaSetValues(fw->filesel.filetext, XtNstring, name, (char *)0);
        add_extra(fw, extra);
        fsel_status = MW_WAITING;

        MwCenter(fsel_pshell);
        XtPopup(fsel_pshell, XtGrabNonexclusive);
        XSetWMProtocols(XtDisplay(fsel_pshell), XtWindow(fsel_pshell),
                        &wm_delete_window, 1);
        fsel_scan(fw);

        XtSetKeyboardFocus(fsel_pshell, fw->filesel.filetext);

        while (fsel_status == MW_WAITING) {
                XEvent event_return;
                XtAppNextEvent(app_context, &event_return);
                XtDispatchEvent(&event_return);
        }

        /* And get rid of the formats */
        XtDestroyWidget(fw->filesel.formatmenu);
        make_dirmenu(fw, "");
        XtDestroyWidget(fw->filesel.dirmenu);
        fw->filesel.dirmenu = None;
        for (i = 0; i < fw->filesel.nextra; i++)
                XtDestroyWidget(fw->filesel.extra[i]);

printf("MwFileselInput() returning path '%s', status %d\n", path, fsel_status);
        return fsel_status;
}

